/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.api.code.authentication.client;


import com.google.common.base.Preconditions;
import no.buypass.api.code.authentication.client.model.UserAuthentication;
import no.buypass.api.code.authentication.client.model.UserCredentials;
import no.buypass.api.code.authentication.client.model.AccessTokenResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.Form;
import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;

/**
 *
 */
public class DefaultAuthenticationClient implements AuthenticationClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(DefaultAuthenticationClient.class);

    private static final GenericType<UserAuthentication> USER_AUTHENTICATION_TYPE = new GenericType<UserAuthentication>(){};

    private static final GenericType<AccessTokenResponse> ACCESS_TOKEN_RESPONSE_TYPE = new GenericType<AccessTokenResponse>(){};

    private static final int DEFAULT_HTTP_PROXY_PORT = 8080;

    private static final String JWT_BEARER_TOKEN_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:jwt-bearer";

    private final WebTarget webTarget;

    private final TokenFactory tokenFactory;

    private final ResponseHandler responseHandler;

    private DefaultAuthenticationClient(Builder builder) {
        webTarget = JerseyClientFactory.createClient().target(builder.serverUrl);
        responseHandler = new DefaultResponseHandler();
        tokenFactory = new TokenFactory(builder.clientCertificate, builder.clientPrivateKey, builder.clientId);
        proxyConfiguration(builder);
    }

    public static Builder builder(String serverUrl, String clientId, X509Certificate clientCertificate, RSAPrivateKey clientPrivateKey) {
        return new Builder(serverUrl, clientId, clientCertificate, clientPrivateKey);
    }

    @Override
    public UserAuthentication authenticateUser(String userName, UserCredentials credentials, String authorization) {
        Response response = webTarget
                .path("user")
                .path(userName)
                .path("authenticate")
                .request()
                .header("Authorization", authorization)
                .post(Entity.json(credentials));

        return responseHandler.processResponse(response, USER_AUTHENTICATION_TYPE);
    }


    @Override
    public UserAuthentication authenticateUser(String userName, UserCredentials userCredentials, String tokenType, String accessToken) {
        return authenticateUser(userName, userCredentials, tokenType + " " + accessToken);
    }

    @Override
    public AccessTokenResponse requestAccessToken() {
        String signedJwt = tokenFactory.createSignedToken();
        return requestAccessToken(JWT_BEARER_TOKEN_GRANT_TYPE, signedJwt);
    }

    private AccessTokenResponse requestAccessToken(String grantType, String signedJwt) {

        Form form = new Form();
        form.param("grant_type", grantType);
        form.param("assertion", signedJwt);

        Response response = webTarget
                .path("auth")
                .path("token")
                .request()
                .post(Entity.entity(form, MediaType.APPLICATION_FORM_URLENCODED_TYPE));

        return responseHandler.processResponse(response, ACCESS_TOKEN_RESPONSE_TYPE);
    }

    private void proxyConfiguration(Builder builder) {
        // HttpUrlConnectorProvider honors the system properties.

        if (builder.proxyUrl != null && !builder.proxyUrl.isEmpty()) {
            try {
                URL url = new URL(builder.proxyUrl);

                String host = url.getHost();
                int port = url.getPort() != -1 ? url.getPort() : DEFAULT_HTTP_PROXY_PORT;

                System.setProperty("http.proxyHost", host);
                System.setProperty("http.proxyPort", String.valueOf(port));
                System.setProperty("https.proxyHost", host);
                System.setProperty("https.proxyPort", String.valueOf(port));

                if (builder.proxyUsername != null && builder.proxyPassword != null) {
                    System.setProperty("http.proxyUser", builder.proxyUsername);
                    System.setProperty("http.proxyPassword", builder.proxyPassword);
                    System.setProperty("https.proxyUser", builder.proxyUsername);
                    System.setProperty("https.proxyPassword", builder.proxyPassword);
                }
                LOGGER.info("Use HTTP proxy {}", builder.proxyUrl);
            } catch (MalformedURLException e) {
                LOGGER.error("Malformed proxyUrl {}", builder.proxyUrl, e);
            }
        }
    }


    public static final class Builder {

        private final String serverUrl;

        private final String clientId;

        private final X509Certificate clientCertificate;

        private final RSAPrivateKey clientPrivateKey;

        private String proxyUrl;

        private String proxyUsername;

        private String proxyPassword;

        public Builder(String serverUrl, String clientId, X509Certificate clientCertificate, RSAPrivateKey clientPrivateKey) {
            this.serverUrl = Preconditions.checkNotNull(serverUrl, "ServerUrl is cannot be null");
            this.clientCertificate = Preconditions.checkNotNull(clientCertificate, "X509Certificate is cannot be null");
            this.clientPrivateKey = Preconditions.checkNotNull(clientPrivateKey, "RSAPrivateKey cannot be null");
            this.clientId = Preconditions.checkNotNull(clientId, "ClientId cannot be null");
        }

        public Builder proxyUrl(String val) {
            proxyUrl = val;
            return this;
        }

        public Builder proxyUsername(String val) {
            proxyUsername = val;
            return this;
        }

        public Builder proxyPassword(String val) {
            proxyPassword = val;
            return this;
        }

        public DefaultAuthenticationClient build() {
            return new DefaultAuthenticationClient(this);
        }
    }
}
