/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.api.code.authentication.client;

import no.buypass.api.code.authentication.client.model.ErrorResponse;
import no.buypass.api.code.authentication.client.exception.AuthenticationClientException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.core.GenericType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;

class DefaultResponseHandler implements ResponseHandler {

    private static final Logger LOG = LoggerFactory.getLogger(DefaultResponseHandler.class);

    public DefaultResponseHandler() {
    }

    @Override
    public <T> T processResponse(Response response, GenericType<T> responseType) {
        return processResponse(response, responseType, Status.OK);
    }

    @Override
    public <T> T processResponse(Response response, GenericType<T> responseType, Status... acceptableStatuses) {
        if (isSuccessStatusCode(response, acceptableStatuses)) {
            return response.readEntity(responseType);
        }
        throw generateException(response);
    }

    private boolean isSuccessStatusCode(Response response, Status... acceptableStatuses) {
        int statusCode = response.getStatus();
        for (Status acceptableStatus : acceptableStatuses) {
            if (acceptableStatus.getStatusCode() == statusCode) {
                return true;
            }
        }
        return false;
    }

    private AuthenticationClientException generateException(Response response) {
        ErrorResponse errorResponse = parseAsErrorResponse(response);
        return new AuthenticationClientException(errorResponse.getDetails(), errorResponse.getCode(), errorResponse.getMessage());
    }

    private ErrorResponse parseAsErrorResponse(Response response) {
        try {
            return response.readEntity(ErrorResponse.class);
        } catch (Exception e) {
            LOG.error(
                    "Not able to parse Response Entity to a ErrorResponse. Got invalid Response Entity in Response " +
                            "'{}'. GENERAL_ERROR",
                    response,
                    e);
            throw new AuthenticationClientException("Was not able to parse the Response body to an ErrorResponse object",
                    e, response.getStatus(), "GENERAL_ERROR" );
        }
    }
}
