/**
 * Copyright 2018 Buypass AS, Oslo, Norway
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package no.buypass.api.code.authentication.client;

import com.google.common.base.Preconditions;
import net.oauth.jsontoken.JsonToken;
import net.oauth.jsontoken.SystemClock;
import net.oauth.jsontoken.crypto.RsaSHA256Signer;
import org.apache.commons.codec.binary.Base64;
import org.joda.time.Duration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SignatureException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;

/**
 *
 */
final class TokenFactory {

    private static final Logger LOG = LoggerFactory.getLogger(TokenFactory.class);

    private static final String AUDIENCE = "https://api.buypass.no";

    private final X509Certificate certificate;
    private final RSAPrivateKey privateKey;
    private final String issuer;

    public TokenFactory(X509Certificate certificate, RSAPrivateKey privateKey, String issuer) {
        this.certificate = Preconditions.checkNotNull(certificate, "X509Certificate is cannot be null");
        this.privateKey = Preconditions.checkNotNull(privateKey, "RSAPrivateKey cannot be null");
        this.issuer = Preconditions.checkNotNull(issuer, "Issuer cannot be null");
    }

    public String createSignedToken() {

        try {
            JsonToken token = createJsonWebToken(certificate, privateKey, AUDIENCE, issuer);
            return token.serializeAndSign();
        } catch (SignatureException e) {
            throw new IllegalStateException("Failed to sign token", e);
        }
    }

    /**
     *
     */
    private JsonToken createJsonWebToken(X509Certificate certificate,
                                        RSAPrivateKey privateKey,
                                        String audience,
                                        String issuer) {
        try {
            RsaSHA256Signer signer = new RsaSHA256Signer(
                    issuer,
                    null,
                    privateKey);

            SystemClock clock = new SystemClock();
            JsonToken token = new JsonToken(signer, clock);

            // X.509 Certificate Thumbprint
            token.getHeader().addProperty("x5t", getThumbPrintSha1(certificate));

            // RFC 7519 4.1.6. "iat" (Issued At) Claim.
            token.setIssuedAt(clock.now());

            // "exp"
            token.setExpiration(clock.now().plus(Duration.standardHours(1)));

            // RFC 7519 4.1.3: "aud" (Audience) Claim. The "aud" (audience) claim identifies the recipients that the JWT is
            // intended for
            token.setAudience(audience);

            return token;
        } catch (InvalidKeyException e) {
            throw new IllegalArgumentException("Illegal Key", e);
        }
    }

    private String getThumbPrintSha1(X509Certificate cert) {
        try {
            MessageDigest md = MessageDigest.getInstance("SHA-1");
            byte[] der = cert.getEncoded();
            md.update(der);
            byte[] digest = md.digest();
            return Base64.encodeBase64URLSafeString(digest);
        } catch (NoSuchAlgorithmException  e) {
            String msg = "SHA-1 algorithm not available.  Fatal (should be in the JDK).";
            LOG.error(msg, e);
            throw new IllegalStateException(msg);
        } catch (CertificateEncodingException e) {
            LOG.error("Certificate encoding error", e);
            throw new IllegalStateException(e.getMessage());
        }
    }
}
